﻿using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Net;
using System.Runtime.InteropServices;
using System.Security;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;

namespace EdgeManage
{
    /// <summary>
    /// This is the Business Logic Layer.  It includes the basic database operations (Create, Retrieve, Update, and
    /// Delete), plus other features to support import, export, copying, moving, sorting, etc..
    /// </summary>
    /// <remarks>
    /// Assumptions: That the USL will disable access to features when the FavoritesDataTable is null
    /// </remarks>
    public class Database
    {
        private ESE_Engine ese;                         // the DAL
        private Favorites.FavoritesDataTable fdt;       // The datatable containing the favorites
        private Guid favBarGUID = Guid.Parse("a62af571-6a95-4ba2-8edd-92a8bb9743f3");
        private Guid rootGUID = Guid.Parse("fddf6d73-3ca3-456b-946a-96b379ad4a44");
        private Stack<Guid> parents;                    // Nest the folder structure during imports
        private Int64 currentOrder;                     // The current order number (for bulk imports)
        private int currentRow;                         // The current row number (for bulk imports)
        private Version edgeVersion;                    // The version of Microsoft Edge
        private SOURCE dataSource;                      // The source of the data
        private string databasePath;                    // The location of the EDB/XML file

        private const Int64 ORDER_INCREMENT = 0x1000000000000;
        private const Int64 ORDER_MASK = 0x7FFF000000000000;

        public enum SOURCE : int
        {
            ESE,            // ordinary ESE where you'd expect to find it
            XML,            // an XML file
            ESE_ReadOnly    // an external ESE file
        };

#region ConstructorsProperties
        private int pfvNewRows;
        private int pfvModifiedRows;
        private int pfvFavorites;
        private int pfvFolders;

        /// <summary>
        /// the location of the order inside the OrderNumber
        /// </summary>
        [StructLayout(LayoutKind.Explicit)]
        private struct OrderOffset
        {
            /*
             * Note: the sort order starts with the 16 high-order bits. But, it's really a lot more complicated than
             * this.  Microsoft uses a "bit crawl" technique for sub-sorts (which I don't follow)
             */
            [FieldOffset(6)]
            public Int16 high;
            [FieldOffset(0)]
            public Int64 big;
        }

        /// <summary>
        /// split the DateUpdated into high and low
        /// </summary>
        [StructLayout(LayoutKind.Explicit)]
        private struct SplitDate
        {
            [FieldOffset(0)]
            public UInt32 low;
            [FieldOffset(4)]
            public UInt32 high;
            [FieldOffset(0)]
            public Int64 big;
        }

        /// <summary>
        /// Default constructor
        /// </summary>
        public Database()
        {
            ese = new ESE_Engine("EdgeManage");
            edgeVersion = Utilities.GetEdgeVersion();
        }

        /// <summary>
        /// Return the number of new rows after an import
        /// </summary>
        public int NewRowCount
        {
            get { return pfvNewRows; }
        }
        /// <summary>
        /// Return the number of modified rows after an import
        /// </summary>
        public int ModifiedRowCount
        {
            get { return pfvModifiedRows; }
        }

        // v2.1 - 5 May 18: Add status strip
        /// <summary>
        /// Return the total number of favorites in the database
        /// </summary>
        public int FavoriteCount
        {
            get { return pfvFavorites; }
        }
        /// <summary>
        /// Return the total number of folders in the database
        /// </summary>
        public int FolderCount
        {
            get { return pfvFolders; }
        }
        #endregion

#region Load&Save

        /// <summary>
        /// Load the Favorites from the ESE database
        /// </summary>
        /// <param name="ds">The type of data</param>
        /// <param name="path">The path to the database (ignored when ds==ESE)</param>
        /// <returns>A fully populated datatable</returns>
        public Favorites.FavoritesDataTable Load(SOURCE ds, string path)
        {
            fdt = null;
            databasePath = "";
            dataSource = ds;

            // v2.1 - 5 May 18: Load an external (readonly) database
            switch (ds)
            {
                case SOURCE.ESE:
                    databasePath = Environment.ExpandEnvironmentVariables(Properties.Settings.Default.DatabasePath);
                    break;
                case SOURCE.ESE_ReadOnly:
                    // v2.1a - 12 May 18: Tweak to be able to read a backup file
                    databasePath = ese.CheckForBackup(path);
                    break;
                case SOURCE.XML:
                    // The XML data source is already in the correct format
                    fdt = new Favorites.FavoritesDataTable();
                    /*
                     * This strange... any valid XML dataset will load without error, so we need another way to
                     * determine if the XML file is corrupt.
                     */
                    fdt.ReadXml(path);
                    if (fdt.Count == 0)
                    {
                        throw new ApplicationException(Properties.Resources.errCorrupt);
                    }
                    fdt.AcceptChanges();
                    currentRow = GetLastRowId() + 1;
                    DoCounts();
                    return fdt;
            }

            // Quick sanity check
            if (!File.Exists(databasePath))
            {
                throw new ApplicationException(Properties.Resources.errMissing);
            }

            // connect to the ESE database and pull the contents of the table
            DataTable dt = ese.GetTable(databasePath, Properties.Settings.Default.TableName, "");

            // create something a bit more civilized
            fdt = new Favorites.FavoritesDataTable();

            // throw an error if the schema doesn't match
            fdt.Merge(dt, false, MissingSchemaAction.Error);

            // The FullPath field is just for the humans...
            // TODO: Should we attempt to keep this current during changes to the rows?
            PopulateFullPath(rootGUID, "\\");

            fdt.AcceptChanges();

            // do a query to find the "standard" GUIDs that we'll need
            bool foundIt = false;
            foreach (Favorites.FavoritesRow fdr in fdt.Select("Title='_Favorites_Bar_'"))
            {
                rootGUID = fdr.ParentId;
                favBarGUID = fdr.ItemId;
                foundIt = true;
                break;
            }

            // another quick sanity check
            if (!foundIt)
            {
                throw new ApplicationException(Properties.Resources.errCorrupt);
            }
            currentRow = GetLastRowId() + 1;

            DoCounts();
            return fdt;
        }

        /// <summary>
        /// Save all cached changes back to the database.  Also creates JSON files to support synchronization if
        /// required
        /// </summary>
        /// <param name="ds">The type of data</param>
        /// <param name="path">The path to the database (ignored when ds==ESE)</param>
        /// <remarks>Note the use of the IsDirty() method to determine if a save is required</remarks>
        public void Save(SOURCE ds, string path)
        {
            if (IsDirty())
            {
                // if ESE database
                if (ds == SOURCE.ESE)
                {
                    // we need to deal with JSON files if synchronization is turned on
                    if (Utilities.IsSyncEnabled())
                    {
                        // are there any changes? (you should not be here otherwise)
                        DataTable tempTable = fdt.GetChanges();
                        if (tempTable != null)
                        {
                            SplitDate sd = new SplitDate();

                            // more civilized version of the table
                            Favorites.FavoritesDataTable tempFav = new Favorites.FavoritesDataTable();
                            tempFav.Merge(tempTable);

                            foreach (Favorites.FavoritesRow fdr in tempFav)
                            {
                                if (fdr.RowState == DataRowState.Deleted || fdr.IsDeleted)
                                {
                                    Guid tempGuid;
                                    if (fdr.RowState == DataRowState.Deleted)
                                    {
                                        tempGuid = (Guid)fdr["ItemId", DataRowVersion.Original];
                                    }
                                    else
                                    {
                                        tempGuid = fdr.ItemId;
                                    }

                                    // delete the existing JSON file
                                    string fileName = String.Format("{0}.json", Path.Combine(
                                        Environment.ExpandEnvironmentVariables(
                                        Properties.Settings.Default.RoamingState), tempGuid.ToString("B")));
                                    if (File.Exists(fileName))
                                    {
                                        /* 
                                         * Note: I'm letting the cleanup routine deal with orphaned icon files
                                         */
                                        File.Delete(fileName);
                                    }
                                }
                                else
                                {
                                    // manually create a JSON file (or overwrite an existing one)
                                    string fileName = String.Format("{0}.json", Path.Combine(
                                        Environment.ExpandEnvironmentVariables(
                                        Properties.Settings.Default.RoamingState), fdr.ItemId.ToString("B")));
                                    using (StreamWriter sw = new StreamWriter(fileName))
                                    {
                                        sw.Write("{\"Collection\":0,\"SchemaVersion\":1,\"ItemId\":\"");
                                        sw.Write(fdr.ItemId.ToString("B"));
                                        sw.Write("\",\"ParentId\":\"");
                                        sw.Write(fdr.ParentId.ToString("B"));
                                        sw.Write("\",\"OrderNumber\":");
                                        // with Edge, this may appear in scientific notation!
                                        sw.Write(fdr.OrderNumber);
                                        sw.Write(",\"IsFolder\":");
                                        sw.Write(fdr.IsFolder.ToString().ToLower());
                                        sw.Write(",\"Title\":\"");
                                        sw.Write(fdr.Title);
                                        sw.Write("\",\"URL\":\"");
                                        if (!fdr.IsURLNull())
                                        {
                                            sw.Write(fdr.URL);
                                        }
                                        sd.big = fdr.DateUpdated;
                                        sw.Write("\",\"DateUpdatedLow\":");
                                        sw.Write(sd.low);
                                        sw.Write(",\"DateUpdatedHigh\":");
                                        sw.Write(sd.high);
                                        sw.Write(",\"FaviconFileContent\":\"");
                                        if (!fdr.IsFaviconFileNull())
                                        {
                                            // convert the icon file into a base64 string
                                            string iconPath = Utilities.GetFullIconPath(fdr.FaviconFile);
                                            if (File.Exists(iconPath))
                                            {
                                                try
                                                {
                                                    byte[] buf = File.ReadAllBytes(iconPath);
                                                    sw.Write(Convert.ToBase64String(buf));
                                                }
                                                catch { }
                                            }
                                        }
                                        sw.Write("\"}");
                                    }
                                }
                            }
                            // TODO: Do we need to trip the ChangeUnitGenerationNeeded value in the registry?
                        }
                    }

                    // v2.2g - 8 Apr 19: Deal with making changes to the IE favorites
                    if (Utilities.IsSyncIE())
                    {
                        /*
                         * There's no practical way to keep track of the changes that need to be applied
                         * to the IE Favorite files.  So, instead, we just whack the whole thing and replace
                         * them all.   We whack everything except for the Links (aka "Favorites Bar") folder
                         * and the desktop.ini files.
                         */
                        DirectoryInfo di = new DirectoryInfo(Environment.GetFolderPath(Environment.SpecialFolder.Favorites));

                        // whack all URLs
                        foreach (FileInfo fi in di.EnumerateFiles("*.url", SearchOption.AllDirectories))
                        {
                            try
                            {
                                fi.Delete();
                            }
                            catch { };
                        }

                        // Whack every folder except for Links (aka "Favorites Bar")
                        foreach (DirectoryInfo diSub in di.EnumerateDirectories())
                        {
                            if (diSub.FullName != Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.Favorites), "Links"))
                            {
                                try
                                {
                                    Directory.Delete(diSub.FullName, true);
                                }
                                catch { };
                            }
                        }

                        // TODO: Figure out how to create the IE registry entry to preserve Edge sort order

                        // put everything back...
                        ExportIE(-1, Environment.GetFolderPath(Environment.SpecialFolder.Favorites));

                        // set the IE sync time property on all rows
                        foreach (Favorites.FavoritesRow fr in fdt)
                        {
                            fr.DateSyncedWithIE = DateTime.Now.ToFileTimeUtc();
                        }
                    }

                    // save the cached DataTable to the database
                    ese.SaveTable(fdt, Environment.ExpandEnvironmentVariables(Properties.Settings.Default.DatabasePath),
                        Properties.Settings.Default.TableName);

                    // update the cached items in the registry?
                    FavBarRegCache();

                    // See if we also need to update the row allocation
                    UpdateRowAllocation();

                    // TODO: Should we update the FullPath field upon save?
                    fdt.AcceptChanges();
                }
                else if (ds == SOURCE.XML)
                {
                    // using an XML data source instead
                    fdt.WriteXml(path);
                    fdt.AcceptChanges();
                }
                DoCounts();
            }
        }
        #endregion

#region CRUD
        /*
         * The CRUD operations 
         */

        /// <summary>
        /// Insert a new row in the cached datatable.  
        /// </summary>
        /// <param name="parentGUID">The GUID of the parent folder</param>
        /// <param name="orderNumber">The order number for this row (or -1 if not known)</param>
        /// <param name="isFolder">Is this entry a favorite or a folder</param>
        /// <param name="nameText">The favorite text as it appears</param>
        /// <param name="urlText">The URL to the favorite</param>
        /// <param name="iconPath">The path to an icon file</param>
        /// <returns>The new row that was just added</returns>
        private Favorites.FavoritesRow Insert(Guid parentGUID, long orderNumber, bool isFolder, string nameText, 
            string urlText, string iconPath)
        {
            // Get the next order number
            if (orderNumber == -1)
            {
                orderNumber = GetNextOrderNumber(parentGUID);
            }

            /*
             * Our home-made rows do not have a HashedUrl.  I haven't been able to figure out how they compute the
             * hash.  But leaving it null doesn't seem to hurt anything.
             */

            // TODO: Figure out the algorithm for HashedUrl!
            Favorites.FavoritesRow row = fdt.NewFavoritesRow();
            row.RowId = currentRow++;   // faster than using GetLastRowId()

            row.IsDeleted = false;
            row.IsFolder = isFolder;
            row.Title = nameText;
            if (!isFolder)
            {
                row.URL = CleanURL(urlText, parentGUID);
            }
            // v2.0d - 27 Dec 17:  Add new DisplayMode field
            row.DisplayMode = true;
            row.DateUpdated = DateTime.Now.ToFileTimeUtc();
            row.ItemId = Guid.NewGuid();
            row.ParentId = parentGUID;
            row.OrderNumber = orderNumber;
            if (iconPath != "")
            {
                row.FaviconFile = iconPath;

                // If this is an XML data source, then fill in FavIconData and then delete the icon file
                if (dataSource == SOURCE.XML)
                {
                    string fileName = Utilities.GetFullIconPath(iconPath);
                    if (File.Exists(fileName))
                    {
                        try
                        {
                            byte[] buf = File.ReadAllBytes(fileName);
                            row.FavIconData = Convert.ToBase64String(buf);
                            File.Delete(fileName);
                        }
                        catch { }
                    }
                }
            }
            fdt.AddFavoritesRow(row);

            return row;
        }

        /// <summary>
        /// Insert a new row in the datatable.  
        /// </summary>
        /// <remarks>Requires the Reorder method to put it where it belongs</remarks>
        /// <param name="folderID">The row number of the parent folder (or -1 if root)</param>
        /// <param name="isFolder">Is this entry a favorite or a folder</param>
        /// <param name="nameText">The favorite text as it appears</param>
        /// <param name="urlText">The URL to the favorite (or "" if not applicable)</param>
        /// <param name="iconPath">The path to an icon file (or "" if not applicable)</param>
        /// <returns>The row number of the new row</returns>
        public int Insert(int folderID, bool isFolder, string nameText, string urlText, string iconPath)
        {
            int answer = 0;
            Guid parent = GetParent(folderID);
            if (parent != Guid.Empty)
            {
                Favorites.FavoritesRow fr = Insert(parent, -1, isFolder, nameText, urlText, iconPath);
                answer = fr.RowId;
            }
            return answer;
        }

        /// <summary>
        /// Update an existing row
        /// </summary>
        /// <param name="rowID">The row number of the record</param>
        /// <param name="nameText">The favorite text (or "" if not applicable)</param>
        /// <param name="urlText">The URL to the favorite (or "" if not applicable)</param>
        public void Update(int rowID, string nameText, string urlText)
        {
            Favorites.FavoritesRow fr = fdt.FindByRowId(rowID);
            if (fr != null)
            {
                // there's only 2 fields that you can edit...
                if (!string.IsNullOrEmpty(nameText))
                {
                    fr.Title = nameText;
                }
                if (!string.IsNullOrEmpty(urlText))
                {
                    fr.URL = CleanURL(urlText, fr.ParentId);
                    fr.SetHashedUrlNull();
                }
                fr.DateUpdated = DateTime.Now.ToFileTimeUtc();
            }
        }

        /// <summary>
        /// Delete an existing row
        /// </summary>
        /// <param name="rowID">The row number of the record</param>
        public void Delete(int rowID)
        {
            // so, this will skip the rowID of -1
            Favorites.FavoritesRow fr = fdt.FindByRowId(rowID);
            if (fr != null)
            {
                // can't delete the special _Favorites_Bar_ folder
                if (!fr.IsDeleted && fr.ItemId != favBarGUID)
                {
                    /*
                     * Check for duplicates... There is a strange rule that allows for duplicate ItemId's.  I have no
                     * idea why... 
                     */
                    // v2.2d - 20 Aug 18: Better check for duplicate ItemId's
                    foreach (Favorites.FavoritesRow frDup in fdt.Select("ItemId='" + fr.ItemId +
                        "' and isDeleted=true"))
                    {
                        // physically remove the old deleted record
                        frDup.Delete();
                    }

                    fr.IsDeleted = true;
                    /*
                     * There is a chance that there already is a deleted item in the same folder with the same URL.
                     * So, we whack the HashedUrl just to be safe.  Yeah, I could look first, but hey...
                     */
                    fr.SetHashedUrlNull();
                    fr.DateUpdated = DateTime.Now.ToFileTimeUtc();
                }
            }

        }

        /// <summary>
        /// Delete a list of rows
        /// </summary>
        /// <param name="rowIDs">A list of row numbers</param>
        public void Delete(List<int> rowIDs)
        {
            foreach (int rowID in rowIDs)
            {
                Favorites.FavoritesRow fr = fdt.FindByRowId(rowID);
                if (fr != null)
                {
                    if (fr.IsFolder)
                    {
                        DeleteTree(rowID);
                        Delete(rowID);
                    }
                    else
                    {
                        Delete(rowID);
                    }
                }
            }
        }

        /// <summary>
        /// Delete a folder recursively (deletes the entire tree)
        /// </summary>
        /// <param name="folderID">The row number of the record</param>
        public void DeleteTree(int folderID)
        {
            foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + GetParent(folderID) + "'"))
            {
                // v2.2d - 20 Aug 18: Better check for duplicate ItemId's
                if (fr.RowState != DataRowState.Detached && fr.RowState != DataRowState.Deleted)
                {
                    if (fr.IsFolder)
                    {
                        // recursive
                        DeleteTree(fr.RowId);
                        Delete(fr.RowId);
                    }
                    else
                    {
                        Delete(fr.RowId);
                    }
                }
            }
        }

        /// <summary>
        /// Mark all of the rows in the DataTable as deleted
        /// </summary>
        /// <returns>the row number of the favorites bar</returns>
        public int DeleteAll()
        {
            int answer = 0;

            // mark everything (except the favorites bar) as deleted
            List<Favorites.FavoritesRow> dups = new List<Favorites.FavoritesRow>();
            foreach (Favorites.FavoritesRow fr in fdt.Select("isDeleted=false"))
            {
                if (fr.ItemId != favBarGUID)
                {
                    // v2.2d - 20 Aug 18: Better check for duplicate ItemId's
                    // search for a duplicate that is already deleted
                    foreach (Favorites.FavoritesRow frDup in fdt.Select("ItemId='" + fr.ItemId +
                        "' and isDeleted=true"))
                    {
                        // mark for physical deletion
                        dups.Add(frDup);
                    }
                    fr.IsDeleted = true;
                    fr.SetHashedUrlNull();
                    fr.DateUpdated = DateTime.Now.ToFileTimeUtc();
                }
                else
                {
                    // the row number of the favorites bar
                    answer = fr.RowId;
                }
            }
            foreach (Favorites.FavoritesRow fr in dups)
            {
                // physically delete the record
                fr.Delete();
            }
            return answer;
        }

        /// <summary>
        /// Undelete a row (by just altering the isDeleted flag)
        /// </summary>
        /// <param name="rowID">The row number of the record</param>
        public void UnDelete(int rowID)
        {
            Favorites.FavoritesRow fr = fdt.FindByRowId(rowID);
            if (fr != null)
            {
                if (fr.IsDeleted)
                {
                    /*
                     * Check for duplicates... There is a strange rule that allows for duplicate ItemId's.  I have
                     * no idea why... 
                     */
                    // v2.2d - x: Better check for duplicate ItemId's
                    foreach (Favorites.FavoritesRow frDup in fdt.Select("ItemId='" + fr.ItemId +
                        "' and isDeleted=false"))
                    {
                        // physically remove the old undeleted record
                        frDup.Delete();
                    }

                    if (!fr.IsURLNull())
                    {
                        string tempURL = fr.URL;
                        // we need to remove the existing URL before running CleanURL
                        fr.URL = "";
                        fr.URL = CleanURL(tempURL, fr.ParentId);
                        fr.SetHashedUrlNull();
                    }
                    fr.IsDeleted = false;
                    fr.DateUpdated = DateTime.Now.ToFileTimeUtc();

                    // v2.0d - 27 Dec 17: Remove junk from the Title
                    int whack = fr.Title.IndexOf("_DeletedBy_");
                    if (whack > 0)
                    {
                        fr.Title = fr.Title.Substring(0, whack);
                    }
                }
            }
        }

        /// <summary>
        /// Undelete a list of rows
        /// </summary>
        /// <param name="rowIDs">A list of row numbers</param>
        public void UnDelete(List<int> rowIDs)
        {
            foreach(int rowID in rowIDs)
            {
                Favorites.FavoritesRow fr = fdt.FindByRowId(rowID);
                if (fr.IsFolder)
                {
                    UnDeleteTree(rowID);
                    UnDelete(rowID);
                }
                else
                {
                    UnDelete(rowID);
                }
            }
        }

        /// <summary>
        /// UnDelete a folder recursively (undeletes the entire tree)
        /// </summary>
        /// <param name="folderID">The row number of the record</param>
        public void UnDeleteTree(int folderID)
        {
            foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + GetParent(folderID) + "'"))
            {
                // v2.2d - 20 Aug 18: Better check for duplicate ItemId's
                if (fr.RowState != DataRowState.Detached && fr.RowState != DataRowState.Deleted)
                {
                    if (fr.IsFolder)
                    {
                        // recursive
                        UnDeleteTree(fr.RowId);
                        UnDelete(fr.RowId);
                    }
                    else
                    {
                        UnDelete(fr.RowId);
                    }
                }
            }
        }

        /// <summary>
        /// Move an item to a new location
        /// </summary>
        /// <remarks>This may be used to change the order within an existing folder</remarks>
        /// <param name="srcID">The row number of the source record</param>
        /// <param name="folderID">The row number of the destination folder</param>
        /// <returns>True if the move was successful</returns>
        public bool Move(int srcID, int folderID)
        {
            bool answer = false;
            Guid destParent = Guid.Empty;

            if (folderID == -1)
            {
                destParent = rootGUID;
            }
            else
            {
                // get the destination parent ID
                Favorites.FavoritesRow frDest = fdt.FindByRowId(folderID);
                if (frDest != null)
                {
                    // The GUI always selects folders for the destID
                    destParent = frDest.ItemId;

                    // quick sanity check
                    if (!frDest.IsFolder)
                    {
                        return false;
                    }
                }
            }

            Favorites.FavoritesRow fr = fdt.FindByRowId(srcID);
            if (fr != null)
            {
                answer = true;

                // Check for "illegal" moves (to a child location)
                if (fr.IsFolder && IsChild(srcID, folderID))
                {
                    answer = false;
                }
                /*
                 * If this within the same folder, do nothing here... Instead, let the Reorder do the work
                 */
                else if (fr.ParentId != destParent)
                {
                    // Change the parent
                    fr.ParentId = destParent;
                    fr.DateUpdated = DateTime.Now.ToFileTimeUtc();

                    // check for duplicate URL
                    if (!fr.IsFolder)
                    {
                        string tempURL = fr.URL;
                        fr.URL = "";
                        // we need to remove the existing URL before running CleanURL
                        fr.URL = CleanURL(tempURL, fr.ParentId);

                        if (tempURL != fr.URL)
                        {
                            fr.SetHashedUrlNull();
                        }
                    }
                    // this is not recursive, since the folder alone "carries" the items below it
                }
                // Note: The OrderNumber is not set here
            }
            return answer;
        }

        /// <summary>
        /// Move a list of items to a new folder
        /// </summary>
        /// <remarks>This preserves the folder structure which is recreated at the destination</remarks>
        /// <param name="rowIDs">The list of rowId to move</param>
        /// <param name="folderID">The destination folder</param>
        /// <param name="sampleID">A rowID at the point of insertion (used get the order number)</param>
        public void Move(List<int> rowIDs, int folderID, int sampleID)
        {
            long nextOrder = ORDER_INCREMENT;

            // get the order number form the "sample" row
            Favorites.FavoritesRow frSample = fdt.FindByRowId(sampleID);
            if (frSample != null)
            {
                nextOrder = frSample.OrderNumber;
            }

            foreach (int rowID in rowIDs)
            {
                Move(rowID, folderID);
                Favorites.FavoritesRow frRow = fdt.FindByRowId(rowID);
                if (frRow != null)
                {
                    // this is just a temporary OrderNumber...
                    frRow.OrderNumber = ++nextOrder;
                }
            }
        }

        // v2.2 - 25 Jun 18: Copy a favorite
        /// <summary>
        /// Copy an item to a new location
        /// </summary>
        /// <param name="srcID">The row number of the item to copy</param>
        /// <param name="folderID">The row number of the destination folder</param>
        /// <returns>The new row number</returns>
        public int Copy(int srcID, int folderID)
        {
            int answer = 0;
            Guid destParent = Guid.Empty;

            if (folderID == -1)
            {
                destParent = rootGUID;
            }
            else
            {
                // get the destination parent ID
                Favorites.FavoritesRow frDest = fdt.FindByRowId(folderID);
                if (frDest != null)
                {
                    destParent = frDest.ItemId;

                    // quick sanity check
                    if (!frDest.IsFolder)
                    {
                        return -1;
                    }
                }
            }

            // get the source record
            Favorites.FavoritesRow frSrc = fdt.FindByRowId(srcID);
            // TODO: Should we skip copying deleted records?
            if (frSrc != null)
            {
                // Check for "illegal" moves (to a child location)
                if (frSrc.IsFolder == false || !IsChild(srcID, folderID))
                {
                    // make a clone of the source record
                    Favorites.FavoritesRow frDest = fdt.NewFavoritesRow();
                    frDest.ItemArray = frSrc.ItemArray.Clone() as object[];

                    // tweak a few things
                    frDest.RowId = ++currentRow;
                    frDest.ItemId = Guid.NewGuid();
                    frDest.ParentId = destParent;
                    frDest.DateUpdated = DateTime.Now.ToFileTimeUtc();
                    fdt.AddFavoritesRow(frDest);

                    answer = frDest.RowId;

                    // check for duplicate URL
                    Favorites.FavoritesRow fr = fdt.FindByRowId(currentRow);
                    if (!fr.IsFolder)
                    {
                        string tempURL = fr.URL;
                        fr.URL = "";
                        // we need to remove the existing URL before running CleanURL
                        fr.URL = CleanURL(tempURL, fr.ParentId);

                        if (tempURL != fr.URL)
                        {
                            fr.SetHashedUrlNull();
                        }
                    }

                    // recursive
                    if (fr.IsFolder)
                    {
                        foreach (Favorites.FavoritesRow frSub in fdt.Select("ParentId='" + frSrc.ItemId + "'"))
                        {
                            Copy(frSub.RowId, frDest.RowId);
                        }
                    }
                }
            }
            return answer;
        }

        // v2.2 - 25 Jun 18: Copy a favorite
        /// <summary>
        /// Copy a list of items to a new folder
        /// </summary>
        /// <param name="rowIDs">The list of rowId to copy</param>
        /// <param name="folderID">The destination folder</param>
        /// <param name="sampleID">A rowID at the point of insertion (used to get the order number)</param>
        /// <returns>A list of new row IDs (currently just used for Unit Testing!)</returns>
        public List<int> Copy(List<int> rowIDs, int folderID, int sampleID)
        {
            List<int> answer = new List<int>();
            long nextOrder = ORDER_INCREMENT;

            // get the order number form the "sample" row
            Favorites.FavoritesRow frSample = fdt.FindByRowId(sampleID);
            if (frSample != null)
            {
                nextOrder = frSample.OrderNumber;
            }

            foreach (int rowID in rowIDs)
            {
                int newRow = Copy(rowID, folderID);
                Favorites.FavoritesRow frRow = fdt.FindByRowId(newRow);
                if (frRow != null)
                {
                    answer.Add(newRow);
                    // this is just a temporary OrderNumber...
                    frRow.OrderNumber = ++nextOrder;
                }
            }
            return answer;
        }

        /// <summary>
        /// Change the order of listing within an existing folder
        /// </summary>
        /// <param name="rowIDs">A list of row number generated from the TrewView</param>
        public void Reorder(List<int> rowIDs)
        {
            /*
             * The Edge browser uses a "secondary sort" technique (to reduce the amount of changes to the database
             * rows?).  This technique seems overly complex and of not much value, so I don't follow it. Yes, that does
             * means that my simple technique will generate a lot more row updates, but hey...
             */
            // TODO: Consider using their technique for the OrderNumber field?

            int index = 1;
            OrderOffset oo = new OrderOffset();
            foreach (int rowID in rowIDs)
            {
                Favorites.FavoritesRow fr = fdt.FindByRowId(rowID);
                if (fr != null)
                {
                    // special case for OrderNumber of the Favorites Bar
                    if (fr.OrderNumber == 1)
                    {
                        continue;
                    }

                    // v2.2 - 25 Jun 18: Remove the bit crawl from the OrderNumber
                    // This will remove any "bit crawl"
                    fr.OrderNumber = (long)((ulong)fr.OrderNumber & 0xFFFF000000000000);
                    oo.big = fr.OrderNumber;
                    if (oo.high != index)
                    {
                        oo.big = 0;
                        oo.high = (Int16)index;
                        fr.OrderNumber = oo.big;
                        fr.DateUpdated = DateTime.Now.ToFileTimeUtc();
                    }
                    index++;
                }
            }
        }
#endregion

#region ImportExport
        /// <summary>
        /// Start the import session
        /// </summary>
        /// <remarks>This creates a "stack" of parent GUIDs to preserve the incoming folder structure</remarks>
        /// <param name="startingGUID">The GUID of the starting folder</param>
        private void ImportStart(Guid startingGUID)
        {
            parents = new Stack<Guid>();
            parents.Push(startingGUID);
            currentOrder = GetNextOrderNumber(startingGUID);

            // v2.0c - 28 Jun 17: build the icon folder if needed
            // verify that the folder structure exists
            string iconfolder = Utilities.GetFullIconPath(@"Data\nouser1\120712-0049\Favorites");
            if (!Directory.Exists(iconfolder))
            {
                try
                {
                    // if it's not there, create it!
                    Directory.CreateDirectory(iconfolder);
                }
                catch { }
            }

            // reset the import counters
            pfvModifiedRows = 0;
            pfvNewRows = 0;
        }

        /// <summary>
        /// Import a folder
        /// </summary>
        /// <remarks>Note the use of the parent GUID stack</remarks>
        /// <param name="rowID">The row number if this folder already exists</param>
        /// <param name="item_Id">The GUID for the ItemID (or Guid.Empty if not used)</param>
        /// <param name="alreadyExists">Does this folder already exist?</param>
        /// <param name="folderName">The name of the folder</param>
        /// <param name="date_Updated">The LastCreated date (or 0 if not used)</param>
        private void ImportFolder(int rowID, Guid item_Id, bool alreadyExists, string folderName, long date_Updated)
        {
            Guid itemGUID;

            if (alreadyExists)
            {
                itemGUID = rootGUID;
                // if it already exists, just update the "Parent stack" below
                Favorites.FavoritesRow fr = fdt.FindByRowId(rowID);
                if (fr != null)
                {
                    // a quick sanity check
                    if (!fr.IsFolder)
                    {
                        return;
                    }

                    // Do we also need to update the Title?
                    if (fr.Title != folderName)
                    {
                        // this only happens if the match is made by ItemId
                        fr.Title = folderName;
                        fr.DateUpdated = DateTime.Now.ToFileTimeUtc();
                    }

                    // v2.2d - 20 Aug 18: Change the ItemId to match
                    if (item_Id != Guid.Empty && fr.ItemId != item_Id)
                    {
                        // do a search and replace on child items
                        foreach (Favorites.FavoritesRow frFolder in fdt.Select("ParentId='" + fr.ItemId + "'"))
                        {
                            frFolder.ParentId = item_Id;
                        }

                        // this only happens if the match is made by Title
                        fr.ItemId = item_Id;
                        // Note: we don't change the DateUpdated
                    }
                    itemGUID = fr.ItemId;
                }
                currentOrder = GetNextOrderNumber(itemGUID);
            }
            else
            {
                // if it doesn't already exist, create it!
                Favorites.FavoritesRow fr = Insert(parents.Peek(), -1, true, folderName, "", "");
                // v2.1d - 4 Jun 18: Add ITEM_ID and LAST_MODIFIED to import/export
                if (date_Updated > 0)
                {
                    fr.DateUpdated = date_Updated;
                }
                if (item_Id != Guid.Empty)
                {
                    fr.ItemId = item_Id;
                }
                itemGUID = fr.ItemId;
                // new folder, so let's start over with a new order number
                currentOrder = ORDER_INCREMENT;
                pfvNewRows++;
            }

            // push the folder's itemGUID on the "Parent stack"
            parents.Push(itemGUID);
        }

        /// <summary>
        /// Import a favorite
        /// </summary>
        /// <remarks>Note the use of the parent GUID stack</remarks>
        /// <param name="rowID">The row number (if it already exists)</param>
        /// <param name="item_Id">The GUID for the ItemID (or Guid.Empty if not used)</param>
        /// <param name="alreadyExists">Does this entry already exist?</param>
        /// <param name="nameText">The favorite name</param>
        /// <param name="urlText">The favorite URL</param>
        /// <param name="iconPath">The path to an icon file (or "" if not available)</param>
        /// <param name="date_Updated">The LastCreated date (or 0 if not used)</param>
        private void ImportFavorite(int rowID, Guid item_Id, bool alreadyExists, string nameText, string urlText, 
            string iconPath, long date_Updated)
        {
            if (alreadyExists)
            {
                // if it already exists, just update the URL and Icon
                Favorites.FavoritesRow fr = fdt.FindByRowId(rowID);
                if (fr != null)
                {
                    // a quick sanity check
                    if (fr.IsFolder)
                    {
                        return;
                    }
                    // Do we need to update the URL?
                    if (fr.URL.TrimEnd("/".ToCharArray()) != urlText.TrimEnd("/".ToCharArray()))
                    {
                        fr.URL = "";
                        // remove the URL above, so the CleanURL won't see it
                        fr.URL = CleanURL(urlText, fr.ParentId);
                        // TODO: Should we use the date_Updated field here?
                        fr.DateUpdated = DateTime.Now.ToFileTimeUtc();
                        // Invalidate the HashedUrl field
                        fr.SetHashedUrlNull();

                        // if the URL has changed, then we also need to swap out the icon
                        if (iconPath != "")
                        {
                            fr.FaviconFile = iconPath;
                        }
                        pfvModifiedRows++;
                    }

                    // Do we also need to update the Title?
                    if (fr.Title != nameText)
                    {
                        // this only happens if the match is made by ItemId
                        fr.Title = nameText;
                        fr.DateUpdated = DateTime.Now.ToFileTimeUtc();
                    }

                    // v2.2d - 20 Aug 18: Change the ItemId to match
                    if (item_Id != Guid.Empty && fr.ItemId != item_Id)
                    {
                        // this only happens if the match is made by Title
                        fr.ItemId = item_Id;
                        // Note: we don't change the DateUpdated
                    }
                }
            }
            else
            {
                Favorites.FavoritesRow fr = Insert(parents.Peek(), currentOrder, false, nameText, urlText, iconPath);
                // v2.1d - 4 Jun 18: Add ITEM_ID and LAST_MODIFIED to import/export
                if (date_Updated > 0)
                {
                    fr.DateUpdated = date_Updated;
                }
                if (item_Id != Guid.Empty)
                {
                    fr.ItemId = item_Id;
                }
                currentOrder += ORDER_INCREMENT;
                pfvNewRows++;
            }
        }

        /// <summary>
        /// Closed down the import session for that folder
        /// </summary>
        private void ImportFolderComplete()
        {
            Guid previous = parents.Pop();
            // v2.2a - 28 Jun 18: Tweak to order numbers during import
            if (parents.Count > 0)
            {
                currentOrder = GetNextOrderNumber(parents.Peek());
            }
            else
            {
                currentOrder = GetNextOrderNumber(previous);
            }
        }

        /// <summary>
        /// Import from an HTML-based bookmarks file
        /// </summary>
        /// <param name="sr">The opened stream that points to the file</param>
        public void ImportHTML(StreamReader sr, bool merge)
        {
            Guid startingGUID = rootGUID;
            string buf;
            ImageConverter imageConv = new ImageConverter();
            bool IsSyncIE = Utilities.IsSyncIE();

            // The 6 patterns for 3 types of lines (start folder, end folder, and favorites)
            Regex folderPattern = new Regex(@"\s*<H3.*?>\s*(.+?)\s*</H3>", RegexOptions.IgnoreCase);
            Regex favoritePattern = new Regex(@"\s*<A HREF=""(.+?)"".*?>(.*?)<\/A>", RegexOptions.IgnoreCase);
            Regex lastModPattern = new Regex(@"\s*LAST_MODIFIED=""(.+?)""", RegexOptions.IgnoreCase);
            Regex itemIdPattern = new Regex(@"\s*ITEM_ID=""(.+?)""", RegexOptions.IgnoreCase);
            Regex iconPattern = new Regex(@"\s*ICON=""data:image/png;base64,(.+?)""", RegexOptions.IgnoreCase);
            Regex endingPattern = new Regex(@"\s*</DL><p>", RegexOptions.IgnoreCase);

            // Support for configurable Merge Import option
            if (!merge)
            {
                // create an "import" folder in the root
                string importFolder = string.Format("Imported_{0:s}", DateTime.Now).Replace(":", "-");
                Favorites.FavoritesRow fr = Insert(rootGUID, -1, true, importFolder, "", "");
                startingGUID = fr.ItemId;
            }

            // start the process
            ImportStart(startingGUID);

            while (sr.Peek() >= 0)
            {
                buf = sr.ReadLine();

                // match the folder line
                Match matchFolder = folderPattern.Match(buf);
                if (matchFolder != Match.Empty)
                {
                    string folderName = matchFolder.Groups[1].Value;
                    bool foundIt = false;
                    int row = 0;

                    // quick sanity check
                    if (String.IsNullOrEmpty(folderName))
                    {
                        continue;
                    }

                    // v2.1d - 4 Jun 18: Add ITEM_ID and LAST_MODIFIED to import/export
                    // read the Last Modified pattern
                    long date_Updated = 0;
                    Match matchMod = lastModPattern.Match(buf);
                    if (matchMod != Match.Empty)
                    {
                        long.TryParse(matchMod.Groups[1].Value, out date_Updated);
                        if (date_Updated != 0)
                        {
                            date_Updated = Utilities.FromEpoch(date_Updated);
                        }
                    }

                    // v2.1d - 4 Jun 18: Add ITEM_ID and LAST_MODIFIED to import/export
                    // read the Item Id pattern
                    Guid item_Id = Guid.Empty;
                    Match matchItem = itemIdPattern.Match(buf);
                    if (matchItem != Match.Empty)
                    {
                        Guid.TryParse(matchItem.Groups[1].Value, out item_Id);
                    }

                    folderName = WebUtility.HtmlDecode(folderName);
                    // v2.2g - 8 Apr 19: Use the IE Sanitize rules
                    if (IsSyncIE)
                    {
                        folderName = Utilities.SanitizeForIE(folderName);
                    }

                    // v2.0c - 28 Jun 17:  translate the Favorites Bar
                    // translate the Favorites Bar
                    if (folderName == "Favorites Bar" && parents.Peek() == rootGUID)
                    {
                        folderName = "_Favorites_Bar_";
                    }

                    // does it already exist?
                    foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + parents.Peek() +
                        "' and Title='" + folderName.Replace("'", "''") + "' and isFolder=true and isDeleted=false"))
                    {
                        row = fr.RowId;
                        foundIt = true;
                        // use only the first match
                        break;
                    }

                    // another duplicate check for the ItemID
                    if (item_Id != Guid.Empty)
                    {
                        foreach (Favorites.FavoritesRow fr in fdt.Select("ItemId='" + item_Id + 
                            "' and isDeleted=false"))
                        {
                            row = fr.RowId;
                            foundIt = true;
                            break;
                        }
                    }

                    // import the folder
                    ImportFolder(row, item_Id, foundIt, folderName, date_Updated);
                    continue;
                }

                // match the Favorites line
                Match matchFavorites = favoritePattern.Match(buf);
                if (matchFavorites.Groups.Count >= 2)
                {
                    string favName = WebUtility.HtmlDecode(matchFavorites.Groups[2].Value);
                    string favURL = matchFavorites.Groups[1].Value;
                    bool foundIt = false;
                    int row = 0;

                    // quick sanity check
                    if (String.IsNullOrEmpty(favURL))
                    {
                        continue;
                    }
                    // v2.2g - 8 Apr 19: Add another sanity check for invalid HTML file format
                    if (String.IsNullOrEmpty(favName))
                    {
                        // FireFox sometimes doesn't include favorite name?!?!
                        favName = favURL;
                    }
                    // v2.2g - 8 Apr 19: Use the IE Sanitize rules
                    if (IsSyncIE)
                    {
                        favName = Utilities.SanitizeForIE(favName);
                    }

                    // v2.1d - 4 Jun 18: Add ITEM_ID and LAST_MODIFIED to import/export
                    // read the Last Modified pattern
                    long date_Updated = 0;
                    Match matchMod = lastModPattern.Match(buf);
                    if (matchMod != Match.Empty)
                    {
                        long.TryParse(matchMod.Groups[1].Value, out date_Updated);
                        if (date_Updated != 0)
                        {
                            date_Updated = Utilities.FromEpoch(date_Updated);
                        }
                    }

                    // v2.1d - 4 Jun 18: Add ITEM_ID and LAST_MODIFIED to import/export
                    // read the Item Id pattern
                    Guid item_Id = Guid.Empty;
                    Match matchItem = itemIdPattern.Match(buf);
                    if (matchItem != Match.Empty)
                    {
                        Guid.TryParse(matchItem.Groups[1].Value, out item_Id);
                    }

                    // does it already exist?
                    foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + parents.Peek() +
                        "' and Title='" + favName.Replace("'", "''") + "' and isFolder=false and isDeleted=false"))
                    {
                        row = fr.RowId;
                        foundIt = true;
                        // use only the first match
                        break;
                    }

                    // another duplicate check for the ItemID (the ItemId has higher precedence)
                    if (item_Id != Guid.Empty)
                    {
                        foreach (Favorites.FavoritesRow fr in fdt.Select("ItemId='" + item_Id +
                            "' and isDeleted=false"))
                        {
                            row = fr.RowId;
                            foundIt = true;
                            break;
                        }
                    }

                    // read the icon file
                    string fileName = "";
                    Match matchIcon = iconPattern.Match(buf);
                    if (matchIcon != Match.Empty)
                    {
                        try
                        {
                            string iconBase64 = matchIcon.Groups[1].Value;
                            byte[] iconBuf = Convert.FromBase64String(iconBase64);

                            // save the icon data
                            fileName = Utilities.GetNewIconPath();
                            File.WriteAllBytes(fileName, iconBuf);

                            // v2.2g - 8 Apr 19: Convert the bitmap to a resized icon
                            // convert to a 24x24 icon if required
                            string tempName = Utilities.GetNewIconPath();
                            if (FavIcon.CreateIcon(fileName, tempName))
                            {
                                fileName = tempName;
                            }
                            /*
                             * Note: If this is importing into a XML data source, then the Insert method (called from 
                             * ImportFavorite) will delete the icon file after it has been read
                             */
                        }
                        catch
                        {
                            fileName = "";
                        }
                    }

                    ImportFavorite(row, item_Id, foundIt, favName, favURL, Utilities.GetIconDataPart(fileName),
                        date_Updated);
                    continue;
                }

                // match the list end
                Match matchEnding = endingPattern.Match(buf);
                if (matchEnding != Match.Empty)
                {
                    ImportFolderComplete();
                }
            }
        }

        /// <summary>
        /// Export favorites into a bookmarks.html file
        /// </summary>
        /// <param name="sw">The opened stream that points to the file</param>
        public void ExportHTML(StreamWriter sw)
        {
            sw.WriteLine("<!DOCTYPE NETSCAPE-Bookmark-file-1>");
            sw.WriteLine("<!-- This is an automatically generated file.");
            sw.WriteLine("     It will be read and overwritten.");
            sw.WriteLine("     Do Not Edit! -->");
            sw.WriteLine("<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=UTF-8\">");
            sw.WriteLine("<TITLE>Bookmarks</TITLE>");
            sw.WriteLine("<H1>Bookmarks</H1>");
            sw.WriteLine("<DL><p>");

            ExportHTMLRecursive(sw, -1, 4);
            sw.WriteLine("</DL><p>");
        }

        /// <summary>
        /// Export favorites into a bookmarks.html file
        /// </summary>
        /// <param name="sw">The opened stream that points to the file</param>
        /// <param name="folderID">The row number of the starting folder (or -1 for root)</param>
        /// <param name="level">The indentation level</param>
        private void ExportHTMLRecursive(StreamWriter sw, int folderID, int level)
        {
            // the amount of indentation in the HTML file
            string indent = new string(' ', level);
            string tempTitle;

            foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + GetParent(folderID) + 
                "' and isDeleted=False", "OrderNumber ASC, Title ASC"))
            {
                // is this a folder?
                if (fr.IsFolder)
                {
                    tempTitle = fr.Title;
                    // translate the Favorites Bar
                    if (tempTitle == "_Favorites_Bar_" && fr.ParentId == rootGUID)
                    {
                        tempTitle = "Favorites Bar";
                    }

                    // v2.1d - 4 Jun 18: Add ITEM_ID and LAST_MODIFIED to import/export
                    // v2.2c - 17 Jul 18: Use SecurityElement.Escape for XML encoding
                    // create a folder
                    sw.WriteLine("{0}<DT><H3 FOLDED ITEM_ID=\"{1}\" LAST_MODIFIED=\"{2}\" >{3}</H3></DT>", indent, 
                        fr.ItemId.ToString("B").ToUpper(), Utilities.ToEpoch(fr.DateUpdated), 
                        SecurityElement.Escape(tempTitle));
                    sw.WriteLine("{0}<DL><p>", indent);

                    // recursively call this routine
                    level += 4;
                    ExportHTMLRecursive(sw, fr.RowId, level);
                    level -= 4;
                    sw.WriteLine("{0}</DL><p>", indent);
                }
                else
                {
                    string tempURL = fr.URL;
                    // v2.2e - 18 Nov 18: Check for invalid URLs already in the database
                    if (Uri.TryCreate(tempURL, UriKind.Absolute, out Uri tempUri))
                    {
                        if (tempUri.Scheme == "http" || tempUri.Scheme == "https")
                        {
                            // in theory, we shouldn't need to clean the URLs
                            tempURL = tempUri.AbsoluteUri;
                        }
                    }

                    // check to see if we have an icon
                    if (!fr.IsFaviconFileNull())
                    {
                        string iconPath = Utilities.GetFullIconPath(fr.FaviconFile);
                        // v2.2 - 25 Jun 18: Find the favicon folder when using an external database
                        if (dataSource == SOURCE.ESE_ReadOnly)
                        {
                            // use an alternate path to find the favicons
                            if (!File.Exists(iconPath))
                            {
                                iconPath = Utilities.GetFullIconPath(databasePath, fr.FaviconFile);
                            }
                        }
                        if (File.Exists(iconPath))
                        {
                            try
                            {
                                byte[] buf = File.ReadAllBytes(iconPath);
                                // v2.1d - 4 Jun 18: Add ITEM_ID and LAST_MODIFIED to import/export
                                // v2.2c - 17 Jul 18: Use SecurityElement.Escape for XML encoding
                                sw.WriteLine("{0}<DT><A HREF=\"{1}\" ITEM_ID=\"{2}\" LAST_MODIFIED=\"{3}\" " +
                                    "ICON=\"data:image/png;base64,{4}\">{5}</A>",
                                    indent, tempURL, fr.ItemId.ToString("B").ToUpper(), 
                                    Utilities.ToEpoch(fr.DateUpdated), Convert.ToBase64String(buf),
                                    SecurityElement.Escape(fr.Title));
                                continue;
                            }
                            catch { }
                        }
                    }
                    // v2.1d - 4 Jun 18: Add ITEM_ID and LAST_MODIFIED to import/export
                    // v2.2c - 17 Jul 18: Use SecurityElement.Escape for XML encoding
                    sw.WriteLine("{0}<DT><A HREF=\"{1}\" ITEM_ID=\"{2}\" LAST_MODIFIED=\"{3}\">{4}</A>", indent, 
                        tempURL, fr.ItemId.ToString("B").ToUpper(), Utilities.ToEpoch(fr.DateUpdated),
                        SecurityElement.Escape(fr.Title));
                }
            }
        }

        /// <summary>
        /// Import from Internet Explorer
        /// </summary>
        /// <param name="pathCurrent">The starting folder</param>
        /// <param name="merge">Show we merge the incoming items?</param>
        public void ImportIE(string pathCurrent, bool merge)
        {
            Guid startingGUID = rootGUID;

            // Support for configurable Merge Import option
            if (!merge)
            {
                // create an "import" folder in the root
                string importFolder = string.Format("Imported_{0:s}", DateTime.Now).Replace(":", "-");
                Favorites.FavoritesRow fr = Insert(rootGUID, -1, true, importFolder, "", "");
                startingGUID = fr.ItemId;
            }

            ImportStart(startingGUID);
            ImportIERecursive(pathCurrent);
        }

        /// <summary>
        /// Import from Internet Explorer
        /// </summary>
        /// <param name="pathCurrent">The path to the favorites folder</param>
        private void ImportIERecursive(string pathCurrent)
        {
            bool IsSyncIE = Utilities.IsSyncIE();

            // do the folders
            foreach (string folderPath in Directory.EnumerateDirectories(pathCurrent))
            {
                string folderName = Path.GetFileName(folderPath);

                // translate the Favorites Bar
                if (folderName == "Links" && parents.Peek() == rootGUID)
                {
                    folderName = "_Favorites_Bar_";
                }

                // v2.2g - 8 Apr 19: Use the IE Sanitize rules
                if (IsSyncIE)
                {
                    folderName = Utilities.SanitizeForIE(folderName);
                }
                bool foundIt = false;
                int row = 0;

                // does it already exist?
                foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + parents.Peek() +
                    "' and Title='" + folderName.Replace("'", "''") + "' and isFolder=true and isDeleted=false"))
                {
                    row = fr.RowId;
                    foundIt = true;
                    // use only the first match
                    break;
                }

                // if not found, then build it
                ImportFolder(row, Guid.Empty, foundIt, folderName, 0);

                // recursive
                ImportIERecursive(folderPath);
                ImportFolderComplete();
            }

            // do the favorite files
            foreach (string fileName in Directory.EnumerateFiles(pathCurrent, "*.url"))
            {
                string favName = Path.GetFileNameWithoutExtension(fileName);
                string favURL = "";
                string buf;
                bool foundIt = false;
                int row = 0;

                // v2.2g - 8 Apr 19: Use the IE Sanitize rules
                if (IsSyncIE)
                {
                    favName = Utilities.SanitizeForIE(favName);
                }

                // does it already exist?
                foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + parents.Peek() + 
                    "' and Title='" + favName.Replace("'", "''") + "' and isFolder=false and isDeleted=false"))
                {
                    row = fr.RowId;
                    foundIt = true;
                    // use only the first match
                    break;
                }

                // read the URL
                try
                {
                    using (StreamReader sr = new StreamReader(fileName))
                    {
                        while (sr.Peek() >= 0)
                        {
                            buf = sr.ReadLine();
                            if (buf.StartsWith("URL=", StringComparison.CurrentCultureIgnoreCase))
                            {
                                favURL = buf.Remove(0, 4);
                                break;
                            }
                            /*
                             * I don't attempt to retrieve the icon from the shortcut... that would take too long
                             */
                        }
                    }
                }
                catch { }

                // quick sanity check
                if (String.IsNullOrEmpty(favName) || String.IsNullOrEmpty(favURL))
                {
                    continue;
                }

                ImportFavorite(row, Guid.Empty, foundIt, favName, favURL, "", 0);
            }
        }

        /// <summary>
        /// Export Edge Favorites to Internet Explorer
        /// </summary>
        /// <param name="folderID">The row number of the starting folder (or -1 for root)</param>
        /// <param name="pathCurrent">The path to the favorites folder</param>
        public void ExportIE(int folderID, string pathCurrent)
        {
            StringBuilder sb = new StringBuilder();
            string buf, fileName;
            bool isSyncIE = Utilities.IsSyncIE();

            // there's no need to honor the sort here, since IE doesn't use that technique
            foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + GetParent(folderID) + "'"))
            {
                // skip deleted items
                if (fr.IsDeleted)
                {
                    continue;
                }

                // if it is a folder
                if (fr.IsFolder)
                {
                    string holdPath = pathCurrent;

                    // make sure it's a valid path for the file system
                    if (isSyncIE)
                    {
                        // v2.2g - 8 Apr 19: Use the IE Sanitize rules
                        pathCurrent = Path.Combine(pathCurrent, Utilities.SanitizeForIE(fr.Title));
                    }
                    else
                    {
                        sb.Clear();
                        sb.Append(fr.Title);
                        foreach (char c in Path.GetInvalidPathChars())
                        {
                            sb.Replace(c.ToString(), " ");
                        }
                        // v2.1 - 5 May 18: Trim spaces
                        pathCurrent = Path.Combine(pathCurrent, sb.ToString().Trim());
                    }

                    // translate the Favorites Bar
                    if (pathCurrent.EndsWith("_Favorites_Bar_") && fr.ParentId == rootGUID)
                    {
                        pathCurrent = pathCurrent.Replace("_Favorites_Bar_", "Links");
                    }

                    // do we need to create this directory?
                    if (!Directory.Exists(pathCurrent))
                    {
                        Directory.CreateDirectory(pathCurrent);
                    }

                    // recursive call
                    ExportIE(fr.RowId, pathCurrent);
                    pathCurrent = holdPath;
                }
                else
                {
                    // make sure it's a valid filename for the file system
                    if (isSyncIE)
                    {
                        // v2.2g - 8 Apr 19: Use the IE Sanitize rules
                        fileName = Path.Combine(pathCurrent, Utilities.SanitizeForIE(fr.Title) + ".url");
                    }
                    else
                    {
                        sb.Clear();
                        sb.Append(fr.Title);
                        foreach (char c in Path.GetInvalidFileNameChars())
                        {
                            sb.Replace(c.ToString(), " ");
                        }
                        // v2.1 - 5 May 18: Trim spaces
                        fileName = Path.Combine(pathCurrent, sb.ToString().Trim() + ".url");
                    }

                    string tempURL = fr.URL;
                    // v2.2e - 18 Nov 18: Check for invalid URLs already in the database
                    if (Uri.TryCreate(tempURL, UriKind.Absolute, out Uri tempUri))
                    {
                        if (tempUri.Scheme == "http" || tempUri.Scheme == "https")
                        {
                            // in theory, we shouldn't need to clean the URLs
                            tempURL = tempUri.AbsoluteUri;
                        }
                    }

                    // does the URL file exist?
                    if (!File.Exists(fileName))
                    {
                        // create it
                        using (StreamWriter sw = new StreamWriter(fileName))
                        {
                            sw.WriteLine("[{000214A0-0000-0000-C000-000000000046}]");
                            sw.WriteLine("Prop3=19,2");
                            sw.WriteLine("[InternetShortcut]");
                            sw.WriteLine("IDList=");
                            sw.WriteLine("URL={0}", tempURL);
                            sw.WriteLine("IconFile={0}", tempURL);
                            sw.WriteLine("IconIndex=1");
                        }
                    }
                    else
                    {
                        sb.Clear();

                        // read the existing content first
                        using (StreamReader sr = new StreamReader(fileName))
                        {
                            while (sr.Peek() >= 0)
                            {
                                buf = sr.ReadLine();
                                if (buf.StartsWith("URL=", StringComparison.CurrentCultureIgnoreCase))
                                {
                                    // just change this one line...
                                    buf = String.Format("URL={0}", tempURL);
                                }
                                sb.AppendLine(buf);
                            }
                        }

                        // now write it back
                        using (StreamWriter sw = new StreamWriter(fileName))
                        {
                            sw.Write(sb.ToString());
                        }
                    }
                }
            }
        }
#endregion

#region Utilities
        /// <summary>
        /// Fill the TreeView with data
        /// </summary>
        /// <param name="showDeleted">Should we show deleted items in the tree?</param>
        /// <param name="tvMain">The treeView control</param>
        public void PopulateTree(bool showDeleted, TreeView tvMain)
        {
            tvMain.BeginUpdate();
            tvMain.Nodes.Clear();

            // start at the top with "root" node
            TreeNode tn = tvMain.Nodes.Add("-1", Properties.Resources.txtTop, fmMain.FOLDER_INDEX, fmMain.FOLDER_INDEX);

            // start at the "root" GUID
            RecursivePopulate(rootGUID, showDeleted, tn);

            tvMain.SelectedNode = tn;
            tn.Expand();
            tvMain.EndUpdate();
        }

        /// <summary>
        /// Recursively fill a node in the TreeView
        /// </summary>
        /// <param name="ParentID">The starting database "folder"</param>
        /// <param name="showDeleted">Should we also include deleted items?</param>
        /// <param name="tn">The current node in the TreeView</param>
        private void RecursivePopulate(Guid ParentID, bool showDeleted, TreeNode tn)
        {
            string query;

            if (showDeleted)
            {
                query = "ParentId='" + ParentID + "'";
            }
            else
            {
                query = "ParentId='" + ParentID + "' and IsDeleted=false";
            }
            foreach (Favorites.FavoritesRow fr in fdt.Select(query, "OrderNumber ASC, Title ASC"))
            {
                if (fr.IsFolder)
                {
                    // create a folder node
                    TreeNode previousNode = tn;
                    tn = tn.Nodes.Add(fr.RowId.ToString(), fr.Title, fmMain.FOLDER_INDEX, fmMain.FOLDER_INDEX);
                    if (showDeleted && fr.IsDeleted)
                    {
                        tn.NodeFont = Utilities.deletedFont;
                        tn.BackColor = Color.Yellow;
                    }
                    // recursive
                    RecursivePopulate(fr.ItemId, showDeleted, tn);
                    tn = previousNode;
                }
                else
                {
                    // create a favorite node
                    TreeNode n = tn.Nodes.Add(fr.RowId.ToString(), fr.Title, fmMain.FAVORITE_INDEX, 
                        fmMain.FAVORITE_INDEX);
                    if (showDeleted && fr.IsDeleted)
                    {
                        n.NodeFont = Utilities.deletedFont;
                        n.BackColor = Color.Yellow;
                    }
                }
            }
        }

        /// <summary>
        /// Retrieve the URL property of a record
        /// </summary>
        /// <param name="rowID">The row number of the record</param>
        /// <returns>The URL text</returns>
        public string GetURL(int rowID)
        {
            string answer = "";

            Favorites.FavoritesRow fr = fdt.FindByRowId(rowID);
            if (fr != null && !fr.IsURLNull())
            {
                answer = fr.URL;
            }
            return answer;
        }

        /// <summary>
        /// Retrieve the Icon path of a record
        /// </summary>
        /// <param name="rowID">The row number of the record</param>
        /// <returns>The full icon path</returns>
        public string GetIcon(int rowID)
        {
            string answer = "";

            Favorites.FavoritesRow fr = fdt.FindByRowId(rowID);
            if (fr != null && !fr.IsFaviconFileNull())
            {
                answer = Utilities.GetFullIconPath(fr.FaviconFile);
            }
            return answer;
        }

        /// <summary>
        /// refresh a single favorite icon
        /// </summary>
        /// <param name="rowId">The row number of the record</param>
        /// <returns>True if it was successful</returns>
        public bool RefreshIcon(int rowId)
        {
            bool answer = false;
            string fileName;
            bool generated = false;

            Favorites.FavoritesRow fr = fdt.FindByRowId(rowId);
            if (fr != null)
            {
                // do we already have an icon file?
                if (fr.IsFaviconFileNull())
                {
                    fileName = Utilities.GetNewIconPath();
                    generated = true;
                }
                else
                {
                    fileName = Utilities.GetFullIconPath(fr.FaviconFile);
                }

                // get and save the icon
                if (FavIcon.DownloadFaviconFile(fr.URL, fileName))
                {
                    answer = true;
                    // update the database if we generated the path
                    if (generated)
                    {
                        fr.FaviconFile = Utilities.GetIconDataPart(fileName);
                        fr.DateUpdated = DateTime.Now.ToFileTimeUtc();
                    }
                    // if this is an XML data source
                    if (dataSource == SOURCE.XML)
                    {
                        try
                        {
                            byte[] buf = File.ReadAllBytes(fileName);
                            fr.FavIconData = Convert.ToBase64String(buf);
                            // delete the file... we don't need it anymore
                            File.Delete(fileName);
                        }
                        catch { }
                    }
                }
            }
            return answer;
        }

        /// <summary>
        /// Set an icon based upon a provided graphic file
        /// </summary>
        /// <param name="rowID">The row number of the record</param>
        /// <param name="graphicFile">The path to the graphic file</param>
        /// <returns>True if successful</returns>
        public bool SetIcon(int rowID, string graphicFile)
        {
            bool answer = false;
            // get a new unique icon name
            string fileName = Utilities.GetNewIconPath();

            /*
             * We ignore any existing icon file.   Any orphaned icon files will be deleted by the cleanup method
             */

            // do the conversion to a 24x24 icon and store the results in the filename
            if (FavIcon.CreateIcon(graphicFile, fileName))
            {
                // save the changes to the database
                Favorites.FavoritesRow fr = fdt.FindByRowId(rowID);
                if (fr != null)
                {
                    fr.FaviconFile = Utilities.GetIconDataPart(fileName);
                    fr.DateUpdated = DateTime.Now.ToFileTimeUtc();
                    // if this is an XML data source
                    if (dataSource == SOURCE.XML)
                    {
                        try
                        {
                            byte[] buf = File.ReadAllBytes(fileName);
                            fr.FavIconData = Convert.ToBase64String(buf);
                            // delete the file... we don't need it anymore
                            File.Delete(fileName);
                        } catch { }
                    }
                    answer = true;
                }
            }
            return answer;
        }

        /// <summary>
        /// Sort the entries in folder (is not recursive)
        /// </summary>
        /// <param name="folderID">The row number of the starting folder</param>
        public void Sort(int folderID)
        {
            int index = 1;
            OrderOffset oo = new OrderOffset();

            foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + GetParent(folderID) + "'", 
                "isFolder DESC, Title ASC"))
            {
                // special case for OrderNumber of the Favorites Bar
                if (fr.OrderNumber == 1L)
                {
                    continue;
                }

                // OrderNumber is 1-based
                oo.big = fr.OrderNumber;
                if (oo.high != index)
                {
                    oo.big = 0;
                    oo.high = (Int16)index;
                    fr.OrderNumber = oo.big;
                    fr.DateUpdated = DateTime.Now.ToFileTimeUtc();
                }
                index++;
            }
        }

        /// <summary>
        /// Sort the items in a folder recursively
        /// </summary>
        /// <param name="folderID">The row number of the starting folder</param>
        public void SortRecursive(int folderID)
        {
            // v2.0b - 6 Jun 17:  Sort the current folder first
            Sort(folderID);

            foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + GetParent(folderID) + 
                "' and isFolder=True"))
            {
                Sort(fr.RowId);
                SortRecursive(fr.RowId);
            }
        }

        /// <summary>
        /// Get the last rowId value in the datatable
        /// </summary>
        /// <remarks>This is used like an autonumber PK</remarks>
        /// <returns>The last rowId value</returns>
        private int GetLastRowId()
        {
            /*
             * I'm not using the MaxAllocatedRowId field in the RowId table here, but I do alter that table (if needed)
             * after a save operation.
             */
            int answer = 0;
            // find the last rowId (use like an auto number)
            foreach (Favorites.FavoritesRow fr in fdt.Select("", "rowId DESC"))
            {
                answer = fr.RowId;
                break;
            }
            return answer;
        }

        /// <summary>
        /// Does the database have any changes that need to be saved?
        /// </summary>
        /// <returns>True if changes need to be saved</returns>
        public bool IsDirty()
        {
            bool answer = false;
            if (fdt != null)
            {
                answer = (fdt.GetChanges() != null);
            }
            return answer;
        }

        /// <summary>
        /// Alter the URL so that it is not duplicated
        /// </summary>
        /// <param name="inputURL">The original URL</param>
        /// <param name="parentGUID">The "folder" for the current item</param>
        /// <returns>A replacement URL that may have been altered</returns>
        private string CleanURL(string inputURL, Guid parentGUID)
        {
            // v2.2b - 9 Jul 18: More efficient CleanURL technique
            /*
             * Edge doesn't allow any duplicate URLs in the same folder.  This is a silly restriction, so I do a
             * workaround.  I add a harmless HTML "fragment" or "query" to the URL to make it unique.
             */
            string answer = "";

            // quick sanity check
            if (!String.IsNullOrEmpty(inputURL))
            {
                answer = inputURL;

                // is it a valid URL?
                if (Uri.TryCreate(inputURL, UriKind.Absolute, out Uri tempUri))
                {
                    // is it a scheme that we are willing to mess with?
                    if (tempUri.Scheme == "http" || tempUri.Scheme == "https")
                    {
                        // v2.2e - 18 Nov 18: Only sanitize http and https
                        // sanitize the input... lower case, etc..
                        answer = tempUri.AbsoluteUri;

                        // no fragment... no query
                        string searchURL = tempUri.GetLeftPart(UriPartial.Path);
                        string query;

                        // there is a behavioral change with the Anniversary Update
                        if (edgeVersion.Major < 38)
                        {
                            // no duplicate URLs anywhere!
                            query = "IsDeleted=false AND IsFolder=false";
                        }
                        else
                        {
                            // no duplicates within the same folder
                            query = "IsDeleted=false AND IsFolder=false AND ParentId='" + parentGUID + "'";
                        }

                        // make a list of potential matching URLs
                        List<string> matchingURLs = new List<string>();
                        foreach (Favorites.FavoritesRow fr in fdt.Select(query))
                        {
                            // Note: There's no way to escape an embedded %, so can't use the query technique
                            if (fr.URL.StartsWith(searchURL))
                            {
                                matchingURLs.Add(fr.URL);
                            }
                        }
                        
                        // look at each potential duplicate
                        while (matchingURLs.Contains(answer))
                        {
                            UriBuilder ub = new UriBuilder(answer);

                            // Plan A - add a fragment
                            if (ub.Fragment == "")
                            {
                                ub.Fragment = "duplicate=0";
                            }
                            else
                            {
                                // do you already have one of our fragments?
                                Match mFrag = Regex.Match(ub.Fragment, @"#duplicate=(\d+)");
                                if (mFrag.Success)
                                {
                                    int dupCount = int.Parse(mFrag.Groups[1].ToString());
                                    ub.Fragment = "duplicate=" + (++dupCount).ToString();
                                }
                                else
                                {
                                    // Plan B - add a query
                                    if (ub.Query == "")
                                    {
                                        ub.Query = "duplicate=0";
                                    }
                                    else
                                    {
                                        // do you already have one of our queries?
                                        Match mQry = Regex.Match(ub.Query, @"duplicate=(\d+)");
                                        if (mQry.Success)
                                        {
                                            int dupCount = int.Parse(mQry.Groups[1].ToString());
                                            string before = mQry.Value;
                                            string after = mQry.Value.Replace(dupCount.ToString(), 
                                                (++dupCount).ToString());
                                            ub.Query = ub.Query.Replace(before, after).Substring(1);
                                        }
                                        else
                                        {
                                            ub.Query = ub.Query.Substring(1) + "&duplicate=0";
                                        }
                                    }
                                }
                            }
                            // try it again with this URL
                            answer = ub.Uri.AbsoluteUri;
                        }
                    }
                }
            }
            return answer;
        }

        /// <summary>
        /// Update the MaxAllocatedRowId field in the RowId table
        /// </summary>
        private void UpdateRowAllocation()
        {
            /*
             * This fixed a bug reported by a user... but I'm not completely certain in what order the save steps
             * should be taking... should this be first or after the database save???
             */
            UInt32 theirMax = 0;
            UInt32 ourMax = (UInt32)GetLastRowId();

            // get the existing MaxAllocatedRowId value from the RowId table
            DataTable dt = ese.GetTable(Environment.ExpandEnvironmentVariables(
                Properties.Settings.Default.DatabasePath), "RowId", "");
            dt.AcceptChanges();

            foreach (DataRow dr in dt.Select("RowId=9"))
            {
                theirMax = (UInt32)dr["MaxAllocatedRowId"];

                // do we need to update this allocation value?
                if (ourMax > theirMax)
                {
                    dr["MaxAllocatedRowId"] = ourMax + 1;
                    ese.SaveTable(dt, Environment.ExpandEnvironmentVariables(Properties.Settings.Default.DatabasePath),
                        "RowId");
                }
                break;
            }
        }

        /// <summary>
        /// Update the items in the favorites bar registry cache
        /// </summary>
        private void FavBarRegCache()
        {
            /*
             * I wouldn't be too surprised if this registry cache disappeared in future versions.  Ha! It did 
             * disappear in Windows 10 1709.
             */
            bool reDoCache = false;

            // first let's see if there are any changes
            foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + favBarGUID + "'", "", 
                DataViewRowState.Added | DataViewRowState.ModifiedCurrent | DataViewRowState.Deleted))
            {
                reDoCache = true;
                break;
            }

            if (reDoCache)
            {
                RegistryKey key = Registry.CurrentUser.OpenSubKey(@"SOFTWARE\Classes\Local Settings\Software\Microsoft\Windows\CurrentVersion\AppContainer\Storage\microsoft.microsoftedge_8wekyb3d8bbwe\MicrosoftEdge\FavOrder\FavBarCache", true);
                if (key != null)
                {
                    // whack all existing entries
                    foreach (string keyName in key.GetSubKeyNames())
                    {
                        key.DeleteSubKey(keyName);
                    }

                    // create new entries
                    int index = 0;
                    foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + favBarGUID + 
                        "' and isDeleted=false", "OrderNumber ASC"))
                    {
                        RegistryKey subKey = key.CreateSubKey(index.ToString());
                        string pathVal = "";
                        if (!fr.IsFaviconFileNull())
                        {
                            pathVal = Utilities.GetFullIconPath(fr.FaviconFile);
                        }
                        subKey.SetValue("faviconPath", pathVal, RegistryValueKind.String);
                        subKey.SetValue("guid", fr.ItemId.ToString("B").ToUpper(), RegistryValueKind.String);
                        subKey.SetValue("isFolder", fr.IsFolder ? 1 : 0, RegistryValueKind.DWord);
                        subKey.SetValue("title", fr.Title, RegistryValueKind.String);
                        string urlVal = "";
                        if (!fr.IsURLNull())
                        {
                            urlVal = fr.URL;
                        }
                        subKey.SetValue("url", urlVal, RegistryValueKind.String);
                        subKey.Close();
                        index++;
                    }
                    key.Close();
                }
            }
        }

        /// <summary>
        /// Fill in the "FullPath" field in the cached datatable
        /// </summary>
        /// <remarks>This field does not exist in the real database</remarks>
        /// <param name="folderID">The row number of the starting folder</param>
        /// <param name="currentPath">The current path</param>
        private void PopulateFullPath(Guid folderID, string currentPath)
        {
            foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + folderID + "'"))
            {
                if (fr.IsFolder)
                {
                    fr.FullPath = currentPath + fr.Title + "\\";
                    // recursive
                    PopulateFullPath(fr.ItemId, fr.FullPath);
                }
                else
                {
                    fr.FullPath = currentPath + fr.Title;
                }
            }
        }

        /// <summary>
        /// Is this GUID a child of the other?
        /// </summary>
        /// <param name="srcRow">The row ID of the "source" to test</param>
        /// <param name="destRow">The row ID of the "destination"</param>
        /// <returns>True if the destination is a child of the source</returns>
        private bool IsChild(int srcRow, int destRow)
        {
            bool answer = false;
            bool broke;

            // quick sanity check...
            if (srcRow == destRow)
            {
                answer = true;
            }
            else if (destRow != -1)
            {
                Favorites.FavoritesRow frSrc = fdt.FindByRowId(srcRow);
                Favorites.FavoritesRow frDest = fdt.FindByRowId(destRow);
                if (frSrc != null && frDest != null)
                {
                    Guid current = frDest.ParentId;
                    do
                    {
                        if (current == frSrc.ItemId)
                        {
                            answer = true;
                            break;
                        }
                        // walk up the tree...
                        broke = true;
                        foreach (Favorites.FavoritesRow fr in fdt.Select("ItemId='" + current + 
                            "' and isDeleted=false"))
                        {
                            current = fr.ParentId;
                            broke = false;
                        }

                    } while (current != rootGUID || !broke);
                }
            }
            return answer;
        }

        /// <summary>
        /// Get the GUID of the parent folder
        /// </summary>
        /// <param name="rowID">The row number of the record (or -1 for root)</param>
        /// <returns>The GUID suitable for searching the parentId field</returns>
        private Guid GetParent(int rowID)
        {
            Guid parent = Guid.Empty;

            // find the parent GUID
            if (rowID == -1)
            {
                parent = rootGUID;
            }
            else
            {
                // find the row given its rowID
                Favorites.FavoritesRow fr = fdt.FindByRowId(rowID);
                if (fr != null)
                {
                    if (fr.IsFolder)
                    {
                        parent = fr.ItemId;
                    }
                    else
                    {
                        parent = fr.ParentId;
                    }
                }
            }
            return parent;
        }

        /// <summary>
        /// Get the next available order number
        /// </summary>
        /// <param name="parentGUID">The parent GUID for the folder</param>
        /// <returns>The next order number for that "folder"</returns>
        private long GetNextOrderNumber(Guid parentGUID)
        {
            long answer = ORDER_INCREMENT;
            foreach (Favorites.FavoritesRow fr in fdt.Select("ParentId='" + parentGUID + "'", "OrderNumber DESC"))
            {
                answer = (fr.OrderNumber & ORDER_MASK) + ORDER_INCREMENT;
                break;
            }
            return answer;
        }

        /// <summary>
        /// Convert a native order number into an integer
        /// </summary>
        /// <param name="order">The native order number</param>
        /// <returns>The integer equivalent</returns>
        public Int16 ToOrderNumber(long order)
        {
            OrderOffset oo = new OrderOffset();
            oo.big = order;
            return oo.high;
        }

        /// <summary>
        /// Get the number of empty icon files
        /// </summary>
        /// <returns>The count of null or invalid FaviconFile</returns>
        public int GetEmptyIconCount()
        {
            int answer = 0;
            if (fdt != null)
            {
                // v2.0c - 28 Jun 17:  Allow both null and invalid icons
                foreach (Favorites.FavoritesRow fdr in fdt.Select("isFolder=false AND isDeleted=false"))
                {
                    if (fdr.IsFaviconFileNull() || !File.Exists(Utilities.GetFullIconPath(fdr.FaviconFile)))
                    {
                        // if the field is null or the icon doesn't exist
                        answer++;
                    }
                }
            }
            return answer;
        }

        /// <summary>
        /// Generate the missing icon files
        /// </summary>
        public void GenerateIcons(object sender, DoWorkEventArgs e)
        {
            string fileName = "";

            BackgroundWorker worker = sender as BackgroundWorker;

            // v2.0c - 28 Jun 17:  Allow both null and invalid icons
            foreach (Favorites.FavoritesRow fdr in fdt.Select("isFolder=false AND isDeleted=false"))
            {
                if (fdr.IsFaviconFileNull() || !File.Exists(Utilities.GetFullIconPath(fdr.FaviconFile)))
                {
                    if ((worker.CancellationPending == true))
                    {
                        e.Cancel = true;
                        break;
                    }
                    else
                    {
                        // generate a new file name (checks for duplicates)
                        fileName = Utilities.GetNewIconPath();

                        /*
                         * Note: These files get generated even if you never save the changes to the database. But
                         * our cleanup routine will delete any orphaned icons
                         */
                        try
                        {
                            // copy the icon file to the proper location
                            if (FavIcon.DownloadFaviconFile(fdr.URL, fileName))
                            {
                                fdr.FaviconFile = Utilities.GetIconDataPart(fileName);
                                // report success
                                worker.ReportProgress(1);
                            }
                            else
                            {
                                // report failure
                                worker.ReportProgress(0);
                            }
                        }
                        catch
                        {
                            // ignore errors and keep going...
                            worker.ReportProgress(0);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Calculate the number of favorites and folders
        /// </summary>
        private void DoCounts()
        {
            // v2.1 - 5 May 18: Add status strip
            pfvFavorites = fdt.Select("isFolder=false AND isDeleted=false").Length;
            pfvFolders = fdt.Select("isFolder=true AND isDeleted=false").Length;
        }

        /// <summary>
        /// Perform some one-time cleanup activities
        /// </summary>
        /// <remarks>This is called asynchronously at start</remarks>
        public void Cleanup()
        {
            // v2.2b - 9 Jul 18: Async cleanup tasks
            // do some cleanup of a previous update
            string setupFile = Path.Combine(Path.GetTempPath(), Path.GetFileName(Properties.Settings.Default.SetupURL));
            if (File.Exists(setupFile))
            {
                /*
                 * Since the upgrade will require that you close this application, there's not really a way to delete
                 * the setup file immediately after the upgrade
                 */
                try
                {
                    File.Delete(setupFile);
                }
                catch { }
            }

            // whack any old temp folders still laying around
            foreach (string oldFolder in Directory.EnumerateDirectories(Environment.ExpandEnvironmentVariables(
                Properties.Settings.Default.TempPath), "EdgeManage.*"))
            {
                try
                {
                    ZetaLongPaths.ZlpIOHelper.DeleteDirectory(oldFolder, true);
                }
                catch { }
            }

            // fix broken files in Roaming State
            foreach (string file in Directory.EnumerateFiles(Environment.ExpandEnvironmentVariables(
                Properties.Settings.Default.RoamingState)))
            {
                /*
                 * This is a silly issue where some files don't get a valid date.  So, I attempt to detect and
                 * fix this error as best as I can.
                 */
                try
                {
                    DateTime temp = File.GetLastWriteTime(file);
                }
                catch
                {
                    File.SetLastWriteTime(file, DateTime.Now);
                }
            }

            // clean up any orphaned favicons
            if (fdt != null)
            {
                // v2.2 - 25 Jun 18: Clean up orphaned favicons
                foreach (string iconFile in Directory.EnumerateFiles(Path.Combine(
                    Environment.ExpandEnvironmentVariables(Properties.Settings.Default.IconPath),
                    @"Data\nouser1\120712-0049\Favorites"), "*.ico"))
                {
                    // is it an orphan?
                    if (fdt.Select("FaviconFile='" + Utilities.GetIconDataPart(iconFile) + "'").Length == 0)
                    {
                        try
                        {
                            File.Delete(iconFile);
                        }
                        catch { }
                    }
                }
            }
        }
        #endregion
    }
}
