﻿using Microsoft.Win32;
using System;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.IO.Compression;
using System.Management;
using System.Security.Principal;
using System.Text;
using System.Xml;

namespace EdgeManage
{
    /// <summary>
    /// Collection of miscellaneous utility functions
    /// </summary>
    public class Utilities
    {
        public static Font deletedFont = new Font(SystemFonts.DefaultFont, FontStyle.Strikeout);

        /// <summary>
        /// Is the Edge "sync your favorites and reading lists" feature enabled
        /// </summary>
        /// <returns>True if it is enabled</returns>
        public static bool IsSyncEnabled()
        {
            bool answer = true;

            // v2.2h - 22 Apr 19: Allow for sync override
            if (!Properties.Settings.Default.SyncOverride)
            {
                // set to -2 if the key is missing
                int syncSettings = -2;
                try
                {
                    // set to -1 if the value missing
                    syncSettings = (int)Registry.GetValue(@"HKEY_CURRENT_USER\SOFTWARE\Classes\Local Settings\Software\Microsoft\Windows\CurrentVersion\AppContainer\Storage\microsoft.microsoftedge_8wekyb3d8bbwe\MicrosoftEdge\Main", "SyncSettings", -1);
                }
                catch { }

                switch (syncSettings)
                {
                    case -2: // The key does not exist
                        answer = false;
                        break;
                    case -1: // The key exists but the value is missing
                             // see if you're using a Microsoft Account
                        WindowsIdentity wi = WindowsIdentity.GetCurrent();
                        WindowsPrincipal wp = new WindowsPrincipal(wi);

                        // This is the SID for the Microsoft Cloud Account Authentication
                        SecurityIdentifier sid = new SecurityIdentifier("S-1-5-64-36");
                        answer = wp.IsInRole(sid);

                        // v2.2d - 20 Aug 18: Detect a hybrid account
                        // yet another test for a hybrid account
                        if (!answer)
                        {
                            int lastSetState = -2;
                            try
                            {
                                // set to -1 if the value missing
                                lastSetState = (int)Registry.GetValue(@"HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\AccountState", "LastSetState", -1);
                            }
                            catch { }
                            if (lastSetState > 0)
                            {
                                answer = true;
                            }
                        }
                        break;
                    case 1: // The value shows it's enabled
                        answer = true;
                        break;
                    case 0: // The value shows that it's disabled
                        answer = false;
                        break;
                    default: // Not likely, but hey...
                        answer = true;
                        break;
                }
            }
            return answer;
        }

        /// <summary>
        /// Generate a new randomly named icon path
        /// </summary>
        /// <returns>The full path to a new icon file</returns>
        public static string GetNewIconPath()
        {
            Random rand = new Random();
            byte[] buf = new byte[8];
            string iconName;
            string fileName = "";

            // generate a new file name (checks for duplicates)
            do
            {
                /*
                 * I can't tell if there is anything specific about these files names, so I'm assuming they are just
                 * random 13 hex digits
                 */
                rand.NextBytes(buf);
                long longRand = BitConverter.ToInt64(buf, 0);
                long num = Math.Abs(longRand % (0xfffffffffffff - 0x1000000000000)) + 0x1000000000000;

                iconName = string.Format(@"Data\nouser1\120712-0049\Favorites\{0:x}_Icon.ico", num);
                fileName = GetFullIconPath(iconName);
            } while (File.Exists(fileName));

            return fileName;
        }

        /// <summary>
        /// Get the full path to the icon file
        /// </summary>
        /// <param name="partialPath">The part of the path that is stored in the database</param>
        /// <returns>The full path of the icon file</returns>
        public static string GetFullIconPath(string partialPath)
        {
            return Path.Combine(Environment.ExpandEnvironmentVariables(Properties.Settings.Default.IconPath),
                partialPath);
        }

        /// <summary>
        /// Get the full path to the icon file
        /// </summary>
        /// <param name="databasePath">the location of the external database</param>
        /// <param name="partialPath">The part of the path that is stored in the database</param>
        /// <returns>The full path of the icon file</returns>
        public static string GetFullIconPath(string databasePath, string partialPath)
        {
            string answer = "";
            if (!string.IsNullOrEmpty(databasePath) && !string.IsNullOrEmpty(partialPath))
            {
                // v2.2 - 25 Jun 18: Find the favicon folder when using an external database
                // walk up the database path 5 folders
                int pos = databasePath.Length - 1;
                for (int count = 0; count < 5; count++)
                {
                    pos = databasePath.LastIndexOf("\\", pos);
                    if (pos <= 0)
                    {
                        break;
                    }
                    pos--;
                }
                if (pos > 0)
                {
                    databasePath = databasePath.Remove(++pos);
                    if (!databasePath.EndsWith("\\"))
                    {
                        databasePath += "\\";
                    }
                }
                answer = Path.Combine(databasePath, partialPath);
            }
            return answer;
        }

        /// <summary>
        /// Get just the part that is stored in the database
        /// </summary>
        /// <param name="fullPath">The full path of the icon file</param>
        /// <returns>The portion of the path that is stored in the database</returns>
        public static string GetIconDataPart(string fullPath)
        {
            string answer = "";
            if (!string.IsNullOrEmpty(fullPath))
            {
                int location = fullPath.IndexOf(@"Data\nouser1\120712-0049\Favorites\");
                if (location > 0)
                {
                    answer = fullPath.Substring(location);
                }
            }
            return answer;
        }

        /// <summary>
        /// Is the Edge application currently running?
        /// </summary>
        /// <returns>True if it is running</returns>
        /// <remarks>Can be easily fooled by saved tabs</remarks>
        public static bool IsEdgeRunning()
        {
            bool answer = false;
            Version ev = GetEdgeVersion();

            // check to see if Edge is running
            foreach (Process p in Process.GetProcessesByName("MicrosoftEdge"))
            {
                if (p.SessionId == Process.GetCurrentProcess().SessionId)
                {
                    // v2.1d - 4 Jun 18: Deal with Edge background tasks
                    if (!String.IsNullOrEmpty(p.MainWindowTitle) && ev.Major >= 42)
                    {
                        // v2.2c - 17 Jul 18: Add another test for detecting the background process
                        int cpCount = 0;
                        foreach (Process cp in Process.GetProcessesByName("MicrosoftEdgeCP"))
                        {
                            if (cp.SessionId == Process.GetCurrentProcess().SessionId)
                            {
                                cpCount++;
                            }
                        }

                        if (cpCount <= 2)
                        {
                            // silently kill Edge if just in the background
                            p.Kill();
                            p.WaitForExit(500);
                            answer = false;
                            break;
                        }
                    }
                    answer = true;
                }
            }
            return answer;
        }

        /// <summary>
        /// Kill off the MicrosoftEdge process to free the exclusive lock on the database
        /// </summary>
        public static void KillEdge()
        {
            foreach (Process p in Process.GetProcessesByName("MicrosoftEdge"))
            {
                if (p.SessionId == Process.GetCurrentProcess().SessionId)
                {
                    p.Kill();
                    p.WaitForExit(500);
                    break;
                }
            }
        }

        /// <summary>
        /// Create a zip file archive
        /// </summary>
        /// <param name="backupPath">The path to the backup folder</param>
        /// <param name="roamingPath">The path to the roaming state folder</param>
        /// <param name="zipPath">The path to the zip file</param>
        public static void CreateZip(string backupPath, string roamingPath, string zipPath)
        {
            using (FileStream fs = new FileStream(zipPath, FileMode.Create, FileAccess.ReadWrite))
            {
                using (ZipArchive zip = new ZipArchive(fs, ZipArchiveMode.Create))
                {
                    // create an entry to hold the version number of Edge
                    ZipArchiveEntry versionEntry = zip.CreateEntry("ArchiveVersion");
                    using (StreamWriter writer = new StreamWriter(versionEntry.Open()))
                    {
                        Version v = GetEdgeVersion();
                        writer.WriteLine(v.ToString());
                    }

                    // the contents of the User folder
                    foreach (string file in Directory.EnumerateFiles(backupPath, "*.*", SearchOption.AllDirectories))
                    {
                        string relativePath = file.Substring(backupPath.Length + 1);
                        // v2.2a - 28 Jun 18: Skip the RACShare folder (it's often too long)
                        if (!relativePath.StartsWith("Default\\RACShare"))
                        {
                            try
                            {
                                zip.CreateEntryFromFile(file, relativePath);
                            }
                            // ignore paths that are too long (they will be the leftovers from conversion)
                            catch (PathTooLongException) { }
                        }
                    }

                    // the contents of the RoamingState folder
                    foreach (string file in Directory.EnumerateFiles(roamingPath))
                    {
                        string relativePath = Path.Combine("RoamingState", Path.GetFileName(file));
                        try
                        {
                            zip.CreateEntryFromFile(file, relativePath);
                        }
                        catch (PathTooLongException) { }
                    }
                }
            }
        }

        /// <summary>
        /// Get the database schema version from the zip file
        /// </summary>
        /// <param name="zipPath">The path to the zip file</param>
        /// <returns>Version number embedded in the zip file (or 0.0 if not present)</returns>
        public static Version GetZipVersion(string zipPath)
        {
            string buf = "0.0.0";
            using (ZipArchive archive = ZipFile.OpenRead(zipPath))
            {
                foreach (ZipArchiveEntry entry in archive.Entries)
                {
                    if (entry.FullName == "ArchiveVersion")
                    {
                        using (StreamReader reader = new StreamReader(entry.Open()))
                        {
                            buf = reader.ReadLine();
                        }
                        break;
                    }
                }
            }
            return new Version(buf);
        }

        /// <summary>
        /// Extract the contents of the Zip archive file 
        /// </summary>
        /// <param name="backupPath">The path of the backup folder</param>
        /// <param name="roamingPath">The path of the roaming state folder</param>
        /// <param name="zipPath">The path to the zip file</param>
        public static void ExtractZip(string backupPath, string roamingPath, string zipPath)
        {
            string roamParent = Path.GetDirectoryName(roamingPath);

            using (ZipArchive archive = ZipFile.OpenRead(zipPath))
            {
                foreach (ZipArchiveEntry entry in archive.Entries)
                {
                    try
                    {
                        if (entry.FullName == "ArchiveVersion")
                        {
                            // ignore the archive version entry
                        }
                        else if (entry.FullName.StartsWith("RoamingState"))
                        {
                            string folder = Path.GetDirectoryName(Path.Combine(roamParent, entry.FullName));
                            if (!Directory.Exists(folder))
                            {
                                Directory.CreateDirectory(folder);
                            }
                            entry.ExtractToFile(Path.Combine(roamParent, entry.FullName));
                        }
                        // v2.2a - 28 Jun 18: Skip the RACShare folder (it's often too long)
                        else if (!entry.FullName.StartsWith("Default\\RACShare"))
                        {
                            string folder = Path.GetDirectoryName(Path.Combine(backupPath, entry.FullName));
                            if (!Directory.Exists(folder))
                            {
                                Directory.CreateDirectory(folder);
                            }
                            entry.ExtractToFile(Path.Combine(backupPath, entry.FullName));
                        }
                    }
                    // ignore paths that are too long (they will be the leftovers from conversion)
                    catch (PathTooLongException) { }
                }
            }
        }
        
        /// <summary>
        /// Get the OS version
        /// </summary>
        /// <returns>A version object (or 0.0 if not successful)</returns>
        public static Version GetOSVersion()
        {
            string buf = "0.0.0";

            // Use Windows Management Instrumentation (WMI) to get the "real" OS version
            ManagementClass osClass = new ManagementClass("Win32_OperatingSystem");
            foreach (ManagementObject queryObj in osClass.GetInstances())
            {
                buf = (string)queryObj.GetPropertyValue("Version");
            }

            return new Version(buf);
        }

        /// <summary>
        /// Get the Edge version number
        /// </summary>
        /// <returns>A version object (or 0.0 if not successful)</returns>
        public static Version GetEdgeVersion()
        {
            string buf = "0.0.0";
            string filePath = Environment.ExpandEnvironmentVariables(Properties.Settings.Default.EdgeManifest);

            if (File.Exists(filePath))
            {
                using (XmlReader reader = XmlReader.Create(filePath))
                {
                    reader.ReadToFollowing("Identity");
                    reader.MoveToAttribute("Version");
                    buf = reader.Value;
                }
            }
            return new Version(buf);
        }

        /// <summary>
        /// Convert a Windows "long date" to a Unix-style "epoch" time
        /// </summary>
        /// <param name="longDate">a windows "long date"</param>
        /// <returns>The resulting Unix-style date</returns>
        public static long ToEpoch(Int64 longDate)
        {
            DateTime dt = DateTime.FromFileTime(longDate).ToLocalTime();
            TimeSpan ts = dt.ToUniversalTime() - new DateTime(1970, 1, 1);
            return (long)ts.TotalSeconds;
        }

        /// <summary>
        /// Convert a Unix-style "epoch" time to a Windows "long date"
        /// </summary>
        /// <param name="longDate">The "epoch" time</param>
        /// <returns>The resulting long date or 0 on error</returns>
        public static long FromEpoch(Int64 longDate)
        {
            long answer = 0;
            try
            {
                DateTime dt = new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc).AddSeconds(longDate).ToLocalTime();
                answer = dt.ToFileTime();
            }
            catch { }
            return answer;
        }

        // v2.2g - 8 Apr 19: These new methods deal with "sync with IE"
        /// <summary>
        /// Detect the "Sync Favorites Between IE and Microsoft Edge" setting
        /// </summary>
        /// <returns>True if the feature is enabled</returns>
        public static bool IsSyncIE()
        {
            bool answer = false;
            using (RegistryKey key = Registry.CurrentUser.OpenSubKey(@"Software\Policies\Microsoft\MicrosoftEdge\Main"))
            {
                if (key != null)
                {
                    if ((int)key.GetValue("SyncFavoritesBetweenIEAndMicrosoftEdge", 0) == 1)
                    {
                        answer = true;
                    }
                }
            }
            return answer;
        }

        /// <summary>
        /// Sanitize the favorite or folder name to meet stricter IE naming requirements
        /// </summary>
        /// <param name="text">Favorite or folder name to check</param>
        /// <returns>sanitized version of the name</returns>
        public static string SanitizeForIE(string text)
        {
            StringBuilder sb = new StringBuilder(text);

            // start with the normal set... then include "dot"
            foreach (char c in Path.GetInvalidFileNameChars())
            {
                sb.Replace(c.ToString(), " ");
            }
            sb.Replace('.', ' ');

            return sb.ToString().Trim();
        }
    }
}
